@@ -6,7 +6,6 @@ from botocore.config import Config
 from botocore.exceptions import ClientError
 from datetime import datetime
 from timestream import DEFAULT_LOGGER, log_it
-from timestream.batch import batch_records, DEFAULT_MAX_BATCH_SIZE, DEFAULT_MIN_COMMON_DIMENSIONS
 from timestream.dimensions import TimestreamDimensions
 from timestream.load_result import TimestreamLoadResult
 from timestream.record import RejectedTimestreamRecord, TimestreamRecord
@@ -18,6 +17,7 @@ DEFAULT_CLIENT_CONFIG = Config(
     max_pool_connections=5000,
     retries={'max_attempts': 10}
 )
+MAX_WRITE_RECORDS = 100 # based on AWS Timestream limits
 
 
 class TimestreamLoader:
@@ -43,6 +43,39 @@ class TimestreamLoader:
             self._logger.warning(f'Overriding TimestreamLoader.__current_time_ms with value {current_time_ms}')
             self.__current_time_ms = current_time_ms
 
+    @log_it
+    def load(
+        self,
+        data: list,
+        capture_time: datetime,
+        source: str = None,
+        upsert: bool = False,
+        ignore_parse_error: bool = True,
+        batch_by: str = 'dimensions'
+    ) -> dict:
+        '''
+        Performs a Timestream load. Validates the input and parses the data into Timestream records.
+        Returns the result of the load after writing to Timestream.
+        '''
+        return self.__load(data, capture_time, source, upsert, ignore_parse_error, False, batch_by).to_dict()
+
+    @log_it
+    def load_preview(
+        self,
+        data: list,
+        capture_time: datetime,
+        source: str = None,
+        upsert: bool = False,
+        ignore_parse_error: bool = True,
+        batch_by: str = 'dimensions'
+    ) -> dict:
+        '''
+        Performs a dry run of a Timestream load, including validating the input values and parsing data
+        into Timestream records.
+        Returns the result of the load before writing to Timestream.
+        '''
+        return self.__load(data, capture_time, source, upsert, ignore_parse_error, True, batch_by).to_dict()
+
     @log_it
     def does_database_exist(
         self,
@@ -127,55 +160,7 @@ class TimestreamLoader:
                 result = False
         return result
 
-    def load(
-        self,
-        data: list,
-        capture_time: datetime,
-        source: str = None,
-        upsert: bool = False,
-        ignore_parse_error: bool = True,
-        max_batch_size: int = DEFAULT_MAX_BATCH_SIZE,
-        min_common_dimensions: int = DEFAULT_MIN_COMMON_DIMENSIONS
-    ) -> dict:
-        '''
-        Performs a Timestream load. Validates the input and parses the data into Timestream records.
-        Returns the result of the load after writing to Timestream.
-        '''
-        return self.__load(
-            data,
-            capture_time,
-            source, upsert,
-            ignore_parse_error,
-            max_batch_size,
-            min_common_dimensions,
-            False,
-        ).to_dict()
-
-    def load_preview(
-        self,
-        data: list,
-        capture_time: datetime,
-        source: str = None,
-        upsert: bool = False,
-        ignore_parse_error: bool = True,
-        max_batch_size: int = DEFAULT_MAX_BATCH_SIZE,
-        min_common_dimensions: int = DEFAULT_MIN_COMMON_DIMENSIONS
-    ) -> dict:
-        '''
-        Performs a dry run of a Timestream load, including validating the input values and parsing data
-        into Timestream records.
-        Returns the result of the load before writing to Timestream.
-        '''
-        return self.__load(
-            data,
-            capture_time,
-            source, upsert,
-            ignore_parse_error,
-            max_batch_size,
-            min_common_dimensions,
-            True,
-        ).to_dict()
-
+    @log_it
     def __load(
         self,
         data: list,
@@ -183,36 +168,47 @@ class TimestreamLoader:
         source: str = None,
         upsert: bool = False,
         ignore_parse_error: bool = True,
-        max_batch_size: int = DEFAULT_MAX_BATCH_SIZE,
-        min_common_dimensions: int = DEFAULT_MIN_COMMON_DIMENSIONS,
-        preview: bool = False
+        preview: bool = False,
+        batch_by: str = 'dimensions'
     ) -> TimestreamLoadResult:
-        self._logger.info(f'Loading data with {len(data)} elements')
-
         # Parse incoming data into TimestreamRecords
         result = self.__parse_data(data, upsert, ignore_parse_error)
         if preview or result.error:
             return result
 
-        # Batch the accepted records by common dimensions
-        batches = batch_records(
-            result.accepted_records,
-            max_batch_size=max_batch_size,
-            min_common_dimensions=min_common_dimensions
-        )
-        assert all(len(b.records) > 0 for b in batches), "Not all batches have records (shouldn't be possible)"
+        assert batch_by in ['dimensions','size','none'], f'Invalid value for "batch_by"'
+        grouped_records, common_dimensions = {}, {}
+        if batch_by == 'dimensions':
+            # Group records by common dimensions using a hash string
+            for r in result.accepted_records:
+                hashstr = r.dimensions.hashstr()
+                common_dimensions[hashstr] = r.dimensions
+                grouped_records[hashstr] = grouped_records.get(hashstr) or []
+                grouped_records[hashstr].append(r)
+        elif batch_by == 'size':
+            # Place all records in the same group
+            grouped_records['all'] = result.accepted_records
+        elif batch_by == 'none':
+            # Each group gets one record
+            grouped_records = {
+                str(i): result.accepted_records[i]
+                for i in range(len(result.accepted_records))
+            }
+
+        # Assert all groups have records
+        assert all(len(records) > 0 for records in grouped_records.values()), "Not all groups have records (shouldn't be possible)"
 
         # Write records in batches to Timestream
         result = TimestreamLoadResult()
-        for b in batches:
-            dimensions = b.common_dimensions
+        for hashstr, records in grouped_records.items():
+            dimensions = common_dimensions.get(hashstr) or TimestreamDimensions()
             if source:
                 dimensions.add_dimension('source', source, overwrite=True)
-            if b.records and upsert:
-                accepted_records, rejected_records = self.__write_upsert_records(dimensions, b.records)
-            elif b.records:
-                accepted_records, rejected_records = self.__write_insert_records(dimensions, b.records, capture_time)
-
+            if records and upsert:
+                accepted_records, rejected_records = self.__write_upsert_records(dimensions, records)
+            elif records:
+                accepted_records, rejected_records = self.__write_insert_records(dimensions, records, capture_time)
+            
             if accepted_records:
                 result.add_accepted_records(accepted_records)
             if rejected_records:
@@ -220,62 +216,73 @@ class TimestreamLoader:
                 result.error = Exception('Some records were rejected by Timestream')
         return result
 
+    @log_it
     def __write_insert_records(
         self,
-        common_dimensions: TimestreamDimensions,
+        dimensions: TimestreamDimensions,
         records: List[TimestreamRecord],
         capture_time: datetime
     ) -> tuple:
-        self._logger.info(f'Writing insert records with {len(records)} records')
         assert capture_time, 'A capture time needs to be provided'
         capture_time_ms = self.__dt_to_ms(capture_time)
-        accepted_records, rejected_records = records, []
-        try:
-            _ = self.__timestream_write_client.write_records(
-                DatabaseName=self.__database_name,
-                TableName=self.__table_name,
-                CommonAttributes={
-                    'Dimensions': common_dimensions.loadable(),
-                    'MeasureValueType': 'DOUBLE',
-                    'Time': str(capture_time_ms),
-                    'Version': 1
-                },
-                Records=[
-                    r.loadable(omit_dimensions=common_dimensions)
-                    for r in records
-                ]
-            )
-        except ClientError as e:
-            self._logger.error(f'Failed to write insert records: {e}')
-            accepted_records, rejected_records = self.__process_client_error(e, records)
-        return accepted_records, rejected_records
+        all_accepted_records, all_rejected_records = [], []
 
+        # Iterate through batches of the max write size
+        for i in range(0, len(records), MAX_WRITE_RECORDS):
+            records_to_write = records[i: min(i + MAX_WRITE_RECORDS, len(records))]
+            accepted_records, rejected_records = records_to_write, []
+            try:
+                _ = self.__timestream_write_client.write_records(
+                    DatabaseName=self.__database_name,
+                    TableName=self.__table_name,
+                    CommonAttributes={
+                        'Dimensions': dimensions.loadable(),
+                        'MeasureValueType': 'DOUBLE',
+                        'Time': str(capture_time_ms),
+                        'Version': 1
+                    },
+                    Records=[r.loadable() for r in records_to_write]
+                )
+            except ClientError as e:
+                self._logger.error(f'Failed to write insert records: {e}')
+                accepted_records, rejected_records = self.__process_client_error(e, records_to_write)
+            
+            all_accepted_records.extend(accepted_records)
+            all_rejected_records.extend(rejected_records)
+        return all_accepted_records, all_rejected_records
+
+    @log_it
     def __write_upsert_records(
         self,
-        common_dimensions: TimestreamDimensions,
+        dimensions: TimestreamDimensions,
         records: List[TimestreamRecord]
     ) -> tuple:
-        self._logger.info(f'Writing upsert records with {len(records)} records')
-        accepted_records, rejected_records = records, []
-        try:
-            _ = self.__timestream_write_client.write_records(
-                DatabaseName=self.__database_name,
-                TableName=self.__table_name,
-                CommonAttributes={
-                    'Dimensions': common_dimensions.loadable(),
-                    'MeasureValueType': 'DOUBLE',
-                    'Version': self.__current_time_ms
-                },
-                Records=[
-                    r.loadable(omit_dimensions=common_dimensions)
-                    for r in records
-                ]
-            )
-        except ClientError as e:
-            self._logger.error(f'Failed to write upsert records: {e}')
-            accepted_records, rejected_records = self.__process_client_error(e, records)
-        return accepted_records, rejected_records
+        all_accepted_records, all_rejected_records = [], []
 
+        # Iterate through batches of the max write size
+        for i in range(0, len(records), MAX_WRITE_RECORDS):
+            records_to_write = records[i: min(i + MAX_WRITE_RECORDS, len(records))]
+            accepted_records, rejected_records = records_to_write, []
+            try:
+                _ = self.__timestream_write_client.write_records(
+                    DatabaseName=self.__database_name,
+                    TableName=self.__table_name,
+                    CommonAttributes={
+                        'Dimensions': dimensions.loadable(),
+                        'MeasureValueType': 'DOUBLE',
+                        'Version': self.__current_time_ms
+                    },
+                    Records=[r.loadable() for r in records_to_write]
+                )
+            except ClientError as e:
+                self._logger.error(f'Failed to write upsert records: {e}')
+                accepted_records, rejected_records = self.__process_client_error(e, records_to_write)
+
+            all_accepted_records.extend(accepted_records)
+            all_rejected_records.extend(rejected_records)
+        return all_accepted_records, all_rejected_records
+
+    @log_it
     def __process_client_error(self, e: ClientError, records: List[TimestreamRecord]) -> tuple:
         '''
         Processes a boto ClientError from a Timestream write operation. Logs error messages
@@ -290,15 +297,16 @@ class TimestreamLoader:
                 error=Exception(rejection_data['Reason'])
             )
             self._logger.error(f'Rejected record:\n{rejected_record}')
-            rejected_records.append(rejected_record)
+            rejected_records.append(rejected_record) 
             rejected_indices.append(rejection_data['RecordIndex'])
-
+        
         accepted_records = [
             records[i] for i in range(len(records))
             if i not in rejected_indices
         ]
         return accepted_records, rejected_records
 
+    @log_it
     def __parse_data(self, data: List[dict], upsert: bool, ignore_parse_error: bool) -> TimestreamLoadResult:
         '''
         Parses the given data and returns a TimestreamLoadResult that contains:
@@ -320,6 +328,7 @@ class TimestreamLoader:
             result.add_accepted_records(records_from_item)
         return result
 
+    @log_it
     def __parse_item(self, item: dict, upsert: bool, ignore_parse_error: bool) -> List[TimestreamRecord]:
         '''
         Parses a single dictionary into Timestream records. A single dictionary may break down into
